import { assert } from "@canvas-js/utils";
import { decodeClock } from "./clock.js";
import { encodeId } from "./ids.js";
import { getAncestorClocks } from "./utils.js";
export class AncestorIndex {
    db;
    static schema = {
        $ancestors: { id: "primary", links: "json" },
    };
    constructor(db) {
        this.db = db;
    }
    async getLinks(id) {
        const record = await this.db.get("$ancestors", id);
        if (record === null) {
            throw new Error(`ancestor links not found for ${id}`);
        }
        return record.links;
    }
    async setLinks(id, links) {
        await this.db.set("$ancestors", { id, links });
    }
    async getAncestors(id, atOrBefore, results = new Set(), visited = new Set()) {
        assert(atOrBefore > 0, "expected atOrBefore > 0");
        const key = encodeId(id);
        const [clock] = decodeClock(key);
        assert(atOrBefore < clock, "expected atOrBefore < clock");
        const index = Math.floor(Math.log2(clock - atOrBefore));
        const links = await this.getLinks(id);
        for (const ancestorId of links[index]) {
            const ancestorKey = encodeId(ancestorId);
            const [ancestorClock] = decodeClock(ancestorKey);
            if (ancestorClock <= atOrBefore) {
                results.add(ancestorId);
            }
            else if (visited.has(ancestorId)) {
                break;
            }
            else {
                visited.add(ancestorId);
                await this.getAncestors(ancestorId, atOrBefore, results, visited);
            }
        }
        return results;
    }
    async isAncestor(id, ancestorId, visited = new Set()) {
        if (id === ancestorId) {
            return true;
        }
        const key = encodeId(id);
        const [clock] = decodeClock(key);
        const ancestorKey = encodeId(ancestorId);
        const [ancestorClock] = decodeClock(ancestorKey);
        if (clock <= ancestorClock) {
            return false;
        }
        const links = await this.getLinks(id);
        const index = Math.floor(Math.log2(clock - ancestorClock));
        for (const id of links[index]) {
            if (visited.has(id)) {
                continue;
            }
            visited.add(id);
            const result = await this.isAncestor(id, ancestorId, visited);
            if (result) {
                return true;
            }
        }
        return false;
    }
    async indexAncestors(id, parentIds) {
        const key = encodeId(id);
        const [clock] = decodeClock(key);
        const ancestorClocks = Array.from(getAncestorClocks(clock));
        const ancestorLinks = new Array(ancestorClocks.length);
        for (const [i, ancestorClock] of ancestorClocks.entries()) {
            if (i === 0) {
                ancestorLinks[i] = parentIds;
            }
            else {
                const links = new Set();
                for (const childId of ancestorLinks[i - 1]) {
                    const childKey = encodeId(childId);
                    const [childClock] = decodeClock(childKey);
                    if (childClock <= ancestorClock) {
                        links.add(childId);
                    }
                    else {
                        assert(childClock <= ancestorClocks[i - 1], "expected childClock <= ancestorClocks[i - 1]");
                        await this.getAncestors(childId, ancestorClock, links);
                    }
                }
                ancestorLinks[i] = Array.from(links).sort();
            }
        }
        await this.setLinks(id, ancestorLinks);
    }
}
