import { logger } from "@libp2p/logger";
import { yamux } from "@chainsafe/libp2p-yamux";
import { select, handle } from "@libp2p/multistream-select";
import WebSocket from "it-ws/web-socket";
import { connect } from "it-ws/client";
import { pipe } from "it-pipe";
import { pushable } from "it-pushable";
import * as lp from "it-length-prefixed";
import { equals } from "uint8arrays";
import { assert } from "@canvas-js/utils";
import * as sync from "@canvas-js/gossiplog/sync";
import { MissingParentError } from "@canvas-js/gossiplog/errors";
import { decodeId, encodeId } from "../ids.js";
import { getPushProtocol, getSyncProtocol, chunk, encodeEvents, decodeEvents } from "../utils.js";
export const factory = yamux({})({ logger: { forComponent: logger } });
export class NetworkClient {
    gossipLog;
    addr;
    log = logger("canvas:network:client");
    pushProtocol;
    syncProtocol;
    muxer;
    duplex;
    sourceURL;
    eventSource = pushable({ objectMode: true });
    constructor(gossipLog, addr) {
        this.gossipLog = gossipLog;
        this.addr = addr;
        this.pushProtocol = getPushProtocol(gossipLog.topic);
        this.syncProtocol = getSyncProtocol(gossipLog.topic);
        this.sourceURL = addr;
        this.duplex = connect(addr, {
        // websocket: [this.gossipLog.topic],
        });
        this.muxer = factory.createStreamMuxer({
            direction: "outbound",
            onIncomingStream: (stream) => {
                handle(stream, [this.pushProtocol, this.syncProtocol], {
                    log: this.log,
                }).then(this.handleProtocolStream);
            },
            onStreamEnd: (stream) => this.log("stream %s closed (%s)", stream.id, stream.protocol),
        });
        this.gossipLog.addEventListener("message", this.handleMessage);
        pipe(this.duplex, this.muxer, chunk, this.duplex).catch((err) => this.log.error(err));
        const ws = this.duplex.socket;
        ws.addEventListener("open", () => {
            gossipLog.dispatchEvent(new CustomEvent("connect", { detail: { peer: this.sourceURL } }));
        });
        ws.addEventListener("close", () => {
            gossipLog.dispatchEvent(new CustomEvent("disconnect", { detail: { peer: this.sourceURL } }));
        });
        this.duplex.connected().then(async () => {
            const eventStream = await this.newStream(this.pushProtocol);
            pipe(this.eventSource, encodeEvents, lp.encode, eventStream.sink).catch((err) => {
                this.log.error(err);
                eventStream.close();
            });
            pipe(eventStream.source, lp.decode, decodeEvents, this.eventSink).catch((err) => {
                this.log.error(err);
                eventStream.close();
            });
            const [_, heads] = await gossipLog.getClock();
            this.push({ update: { heads: heads.map(encodeId) } });
        });
    }
    async newStream(protocol) {
        this.log("opening new %s stream", protocol);
        const muxerStream = await this.muxer.newStream();
        const { stream } = await select(muxerStream, protocol, { log: this.log });
        return stream;
    }
    async close() {
        this.gossipLog.removeEventListener("message", this.handleMessage);
        await this.muxer.close();
        await this.duplex.close();
    }
    isConnected() {
        return this.duplex.socket.readyState === WebSocket.OPEN;
    }
    push(event) {
        this.eventSource.push(event);
    }
    handleMessage = ({ detail: { id, key, value, source } }) => {
        this.log("handling message %s from source %o", id, source);
        if (source === undefined) {
            this.log("pushing %s to %s", id, this.sourceURL);
            this.push({ insert: { key, value } });
        }
    };
    handleProtocolStream = async ({ protocol, stream }) => {
        if (protocol === this.syncProtocol) {
            this.gossipLog.serve((txn) => sync.Server.handleStream(txn, stream)).catch((err) => this.log(err));
        }
        else if (protocol === this.pushProtocol) {
            throw new Error("incoming push stream rejected");
        }
        else {
            throw new Error("unsupported protocol");
        }
    };
    eventSink = async (events) => {
        for await (const event of events) {
            if (event.insert !== undefined) {
                await this.handleInsertEvent(event.insert);
            }
            else if (event.update !== undefined) {
                await this.handleUpdateEvent(event.update);
            }
            else {
                throw new Error("invalid event");
            }
        }
    };
    async handleInsertEvent({ key, value }) {
        const signedMessage = this.gossipLog.decode(value, {
            source: { type: "push", peer: this.sourceURL },
        });
        assert(equals(key, signedMessage.key), "invalid key");
        this.log("handling insert %s", signedMessage.id);
        try {
            await this.gossipLog.insert(signedMessage);
        }
        catch (err) {
            if (err instanceof MissingParentError) {
                this.sync();
                return;
            }
            else {
                throw err;
            }
        }
    }
    async handleUpdateEvent({ heads }) {
        this.log("handling update: %o", heads.map(decodeId));
        const result = await this.gossipLog.tree.read((txn) => {
            for (const key of heads) {
                const leaf = txn.getNode(0, key);
                if (leaf === null) {
                    return key;
                }
            }
            return null;
        });
        if (result !== null) {
            this.sync();
        }
    }
    async sync() {
        this.log("initiating merkle sync with %s", this.sourceURL);
        do {
            let stream;
            try {
                stream = await this.newStream(this.syncProtocol);
            }
            catch (err) {
                this.log.error("failed to open outgoing sync stream: %O", err);
                return;
            }
            this.log("opened outgoing sync stream %s", stream.id);
            // sync.Client
            const client = new sync.Client(stream);
            try {
                const result = await this.gossipLog.sync(client, { peer: this.sourceURL });
                if (result.complete) {
                    break;
                }
                else {
                    continue;
                }
            }
            catch (err) {
                this.log.error("sync failed: %O", err);
                return;
            }
            finally {
                client.end();
                stream.close().then(() => this.log("closed outgoing sync stream %s", stream.id), (err) => this.log.error("error closing sync stream %s: %O", stream.id, err));
            }
        } while (this.isConnected());
    }
}
