import { TopicValidatorResult, } from "@libp2p/interface";
import { GossipSub } from "@chainsafe/libp2p-gossipsub";
import { logger } from "@libp2p/logger";
import * as lp from "it-length-prefixed";
import { pipe } from "it-pipe";
import { pushable } from "it-pushable";
import { equals } from "uint8arrays";
import PQueue from "p-queue";
import { assert } from "@canvas-js/utils";
import * as sync from "@canvas-js/gossiplog/sync";
import { Event } from "@canvas-js/gossiplog/protocols/events";
import { MissingParentError } from "@canvas-js/gossiplog/errors";
import { decodeId, encodeId } from "../ids.js";
import { DEFAULT_PROTOCOL_SELECT_TIMEOUT, MAX_INBOUND_STREAMS, MAX_OUTBOUND_STREAMS, NEGOTIATE_FULLY, } from "../constants.js";
import { decodeEvents, encodeEvents, getPushProtocol, getSyncProtocol } from "../utils.js";
export class GossipLogService {
    components;
    static extractGossipSub(components) {
        const pubsub = components.pubsub;
        assert(pubsub instanceof GossipSub, "expected pubsub instanceof GossipSub");
        return pubsub;
    }
    pubsub;
    messageLog;
    log;
    syncProtocol;
    pushProtocol;
    syncQueue = new PQueue({ concurrency: 1 });
    syncQueuePeers = new Set();
    pushTopology = new Set();
    eventSources = new Map();
    #pushTopologyId = null;
    #started = false;
    constructor(components, { gossipLog }) {
        this.components = components;
        this.log = logger(`canvas:gossiplog:[${gossipLog.topic}]:service`);
        this.pubsub = GossipLogService.extractGossipSub(components);
        this.syncProtocol = getSyncProtocol(gossipLog.topic);
        this.pushProtocol = getPushProtocol(gossipLog.topic);
        this.messageLog = gossipLog;
    }
    beforeStart() {
        this.log("beforeStart");
        this.components.events.addEventListener("connection:open", this.handleConnectionOpen);
        this.components.events.addEventListener("connection:close", this.handleConnectionClose);
        this.pubsub.addEventListener("gossipsub:message", this.handleGossipsubMessage);
        this.pubsub.addEventListener("gossipsub:graft", this.handleGossipsubGraft);
        this.pubsub.addEventListener("gossipsub:prune", this.handleGossipsubPrune);
        this.messageLog.addEventListener("message", this.handleMessage);
        this.messageLog.addEventListener("sync", this.handleSync);
    }
    async start() {
        if (this.#started) {
            return;
        }
        this.log("start");
        this.#started = true;
        /**
         * sync is the protocol for initiating merkle syncs using the Okra index.
         * syncs are initiated in three cases:
         * 1. when a peer receives an orphan message via gossipsub, it schedules
         *    a sync with the gossipsub message's "propagation source"
         * 2. when a peer receives a push update with missing heads, it schedules
         *    a sync with the sender
         * 3. when a peer receives a push insert with missing parents, it schedules
         *    a sync with the sender
         */
        await this.components.registrar.handle(this.syncProtocol, this.handleSyncStream, {
            maxInboundStreams: MAX_INBOUND_STREAMS,
            maxOutboundStreams: MAX_OUTBOUND_STREAMS,
        });
        /**
         * push is the protocol for sending Update and Insert events to specific peers.
         * - Update events carry the sender's set of latest concurrent message ids ("heads")
         * - Insert events carry a signed message as an encoded [key, value] entry
         * Updates are pushed in two cases:
         * 1. once for every new connection, so that the recipient can merkle sync
         *    back if necessary. this happens from both sides of each connection simultaneously.
         * 2. whenever a peer finishes an merkle sync during which it received one or more
         *    missing messages, it pushes its new heads to all of its peers, except for
         *    the sync source.
         */
        await this.components.registrar.handle(this.pushProtocol, this.handlePushStream, {
            maxInboundStreams: MAX_INBOUND_STREAMS,
            maxOutboundStreams: MAX_OUTBOUND_STREAMS,
        });
        this.#pushTopologyId = await this.components.registrar.register(this.pushProtocol, {
            notifyOnLimitedConnection: false,
            onConnect: (peerId, connection) => {
                this.log("connected to %p", peerId);
                this.pushTopology.add(peerId.toString());
                if (connection.direction === "outbound") {
                    this.newStream(connection, this.pushProtocol).then((stream) => this.handlePushStream({ connection, stream }), (err) => this.log.error("failed to open outgoing push stream: %O", err));
                }
            },
            onDisconnect: (peerId) => {
                this.log("disconnected %p", peerId);
                this.pushTopology.delete(peerId.toString());
            },
        });
    }
    afterStart() {
        this.pubsub.subscribe(this.messageLog.topic);
    }
    beforeStop() {
        this.messageLog.removeEventListener("message", this.handleMessage);
        this.messageLog.removeEventListener("sync", this.handleSync);
        this.components.events.removeEventListener("connection:open", this.handleConnectionOpen);
        this.components.events.removeEventListener("connection:close", this.handleConnectionClose);
        this.pubsub.removeEventListener("gossipsub:message", this.handleGossipsubMessage);
        this.pubsub.removeEventListener("gossipsub:graft", this.handleGossipsubGraft);
        this.pubsub.removeEventListener("gossipsub:prune", this.handleGossipsubPrune);
        this.pubsub.unsubscribe(this.messageLog.topic);
    }
    async stop() {
        this.log("stop");
        if (this.#started === false) {
            return;
        }
        this.#started = false;
        for (const [peer, source] of this.eventSources) {
            source.end();
            this.eventSources.delete(peer);
        }
        await this.components.registrar.unhandle(this.syncProtocol);
        await this.components.registrar.unhandle(this.pushProtocol);
        if (this.#pushTopologyId !== null) {
            this.components.registrar.unregister(this.#pushTopologyId);
            this.#pushTopologyId = null;
        }
    }
    async newStream(connection, protocol) {
        assert(connection.status === "open", "connection closed");
        const protocolSelectSignal = AbortSignal.timeout(DEFAULT_PROTOCOL_SELECT_TIMEOUT);
        const stream = await connection.newStream(protocol, {
            negotiateFully: NEGOTIATE_FULLY,
            signal: protocolSelectSignal,
        });
        return stream;
    }
    handleConnectionOpen = ({ detail: connection }) => {
        this.log("connection:open %s %p", connection.id, connection.remotePeer);
        this.messageLog.dispatchEvent(new CustomEvent("connect", { detail: { peer: connection.remotePeer.toString() } }));
    };
    handleConnectionClose = ({ detail: connection }) => {
        this.log("connection:close %s %p", connection.id, connection.remotePeer);
        this.messageLog.dispatchEvent(new CustomEvent("disconnect", { detail: { peer: connection.remotePeer.toString() } }));
    };
    handleGossipsubGraft = ({ detail: { topic, peerId } }) => {
        if (topic !== this.messageLog.topic) {
            return;
        }
        this.log("gossipsub:graft %s", peerId);
    };
    handleGossipsubPrune = ({ detail: { topic, peerId } }) => {
        if (topic !== this.messageLog.topic) {
            return;
        }
        this.log("gossipsub:prune %s", peerId);
    };
    handleMessage = ({ detail: { id, key, value, source } }) => {
        if (!this.#started) {
            return;
        }
        if (source === undefined || source.type === "push") {
            const event = { insert: { key, value } };
            const data = Event.encode(event);
            this.pubsub.publish(this.messageLog.topic, data).then(() => this.log("published %s", id), (err) => this.log.error("failed to publish message %s: %O", id, err));
        }
    };
    handleSync = ({ detail: { messageCount, peer } }) => {
        if (messageCount > 0) {
            this.messageLog.getClock().then(([_, parents]) => {
                const heads = parents.map(encodeId);
                for (const [id, eventSource] of this.eventSources) {
                    if (peer === id) {
                        continue;
                    }
                    else {
                        eventSource.push({ update: { heads } });
                    }
                }
            });
        }
    };
    handleGossipsubMessage = ({ detail: { msgId, propagationSource, msg }, }) => {
        if (msg.topic !== this.messageLog.topic) {
            return;
        }
        const sourceId = propagationSource.toString();
        this.log("received gossipsub message %s via %s", msgId, sourceId);
        let event;
        try {
            event = Event.decode(msg.data);
        }
        catch (err) {
            this.log.error("error decoding gossipsub message: %O", err);
            this.log.trace("rejecting gossipsub message %s", msgId);
            this.pubsub.reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Reject);
            return;
        }
        if (event.insert === undefined) {
            this.log.trace("ignoring gossipsub message %s", msgId);
            this.pubsub.reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Ignore);
            return;
        }
        let signedMessage;
        try {
            signedMessage = this.messageLog.decode(event.insert.value, {
                source: { type: "pubsub", peer: sourceId },
            });
            assert(equals(event.insert.key, signedMessage.key), "invalid key");
        }
        catch (err) {
            this.log.error("invalid message: %O", err);
            this.log.trace("rejecting gossipsub message %s", msgId);
            this.pubsub.reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Reject);
            return;
        }
        this.messageLog.insert(signedMessage).then(() => {
            this.log.trace("accepting gossipsub message %s", msgId);
            this.pubsub.reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Accept);
        }, (err) => {
            if (err instanceof MissingParentError) {
                this.log.trace("ignoring gossipsub message %s", msgId);
                this.pubsub.reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Ignore);
                this.scheduleSync(propagationSource);
            }
            else {
                this.log.error("error inserting message %s: %O", signedMessage.id, err);
                this.log.trace("rejecting gossipsub message %s", msgId);
                this.pubsub.reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Reject);
            }
        });
    };
    getEventSink = (connection, stream) => async (events) => {
        try {
            for await (const event of events) {
                if (event.insert !== undefined) {
                    // TODO: reject insert events from libp2p peers?
                    await this.handleInsertEvent(connection.remotePeer, event.insert);
                }
                else if (event.update !== undefined) {
                    await this.handleUpdateEvent(connection.remotePeer, event.update);
                }
            }
        }
        catch (err) {
            stream.abort(err instanceof Error ? err : new Error(`${err}`));
        }
    };
    async handleInsertEvent(peerId, { key, value }) {
        const signedMessage = this.messageLog.decode(value, {
            source: { type: "push", peer: peerId.toString() },
        });
        assert(equals(key, signedMessage.key), "invalid key");
        try {
            await this.messageLog.insert(signedMessage);
        }
        catch (err) {
            if (err instanceof MissingParentError) {
                this.scheduleSync(peerId);
                return;
            }
            else {
                throw err;
            }
        }
    }
    async handleUpdateEvent(peerId, { heads }) {
        this.log("handling update: %o", heads.map(decodeId));
        const result = await this.messageLog.tree.read((txn) => {
            for (const key of heads) {
                const leaf = txn.getNode(0, key);
                if (leaf === null) {
                    return key;
                }
            }
            return null;
        });
        if (result !== null) {
            this.scheduleSync(peerId);
        }
    }
    handlePushStream = ({ connection, stream }) => {
        const peerId = connection.remotePeer;
        this.log("opened %s push stream %s from peer %p", stream.direction, stream.id, peerId);
        const eventSource = pushable({ objectMode: true });
        this.eventSources.set(peerId.toString(), eventSource);
        const eventSink = this.getEventSink(connection, stream);
        pipe(eventSource, encodeEvents, lp.encode, stream, lp.decode, decodeEvents, eventSink)
            .catch((err) => this.log.error("error in %s push stream %s: %O", stream.direction, stream.id, err))
            .finally(() => {
            eventSource.end();
            this.eventSources.delete(peerId.toString());
            stream.close();
            this.log("closed %s push stream %s from peer %p", stream.direction, stream.id, peerId);
        });
        this.messageLog.getClock().then(([_, heads]) => {
            eventSource.push({ update: { heads: heads.map(encodeId) } });
        });
    };
    handleSyncStream = async ({ connection, stream }) => {
        const peerId = connection.remotePeer;
        this.log("opened incoming sync stream %s from peer %p", stream.id, peerId);
        try {
            await this.messageLog.serve((txn) => sync.Server.handleStream(txn, stream));
            stream.close();
        }
        catch (err) {
            if (err instanceof Error) {
                this.log.error("aborting incoming stream %s from peer %p: %O", stream.id, peerId, err);
                stream.abort(err);
            }
            else {
                this.log.error("aborting incoming stream %s from peer %p: %O", stream.id, peerId, err);
                stream.abort(new Error("internal error"));
            }
        }
        finally {
            this.log("closed incoming sync stream %s from peer %p", stream.id, peerId);
        }
    };
    scheduleSync(peerId) {
        const id = peerId.toString();
        if (this.syncQueuePeers.has(id)) {
            this.log("already queued sync with %p", peerId);
            return;
        }
        this.log("scheduling sync with %p", peerId);
        this.syncQueuePeers.add(id);
        this.syncQueue
            .add(() => this.sync(peerId))
            .catch((err) => this.log.error("sync failed: %O", err))
            .finally(() => this.syncQueuePeers.delete(id));
    }
    async sync(peerId) {
        this.log("initiating sync with %p", peerId);
        const connection = this.components.connectionManager
            .getConnections(peerId)
            .find((connection) => connection.limits === undefined && connection.status === "open");
        if (connection === undefined) {
            this.log("no longer connected to %p", peerId);
            return;
        }
        do {
            let stream;
            try {
                stream = await this.newStream(connection, this.syncProtocol);
            }
            catch (err) {
                this.log.error("failed to open outgoing sync stream: %O", err);
                return;
            }
            this.log("opened outgoing sync stream %s to peer %p", stream.id, peerId);
            const client = new sync.Client(stream);
            try {
                const result = await this.messageLog.sync(client, { peer: peerId.toString() });
                if (result.complete) {
                    break;
                }
                else {
                    continue;
                }
            }
            catch (err) {
                this.log.error("sync failed: %O", err);
                return;
            }
            finally {
                client.end();
                stream.close().then(() => this.log("closed outgoing sync stream %s", stream.id), (err) => this.log.error("error closing sync stream %s: %O", stream.id, err));
            }
        } while (connection.status === "open");
    }
}
export const gossipLogService = (init) => (components) => new GossipLogService(components, init);
